@@ -68,7 +68,6 @@ class Move(metaclass=PoolMeta):
         '''
         pool = Pool()
         Uom = pool.get('product.uom')
-        Currency = pool.get('currency.currency')
 
         total_qty = Uom.compute_qty(self.uom, self.quantity,
             self.product.default_uom, round=False)
@@ -81,24 +80,15 @@ class Move(metaclass=PoolMeta):
         to_save = []
         for move, move_qty in fifo_moves:
             consumed_qty += move_qty
-            if move.from_location.type in {'supplier', 'production'}:
-                with Transaction().set_context(date=move.effective_date):
-                    move_unit_price = Currency.compute(
-                        move.currency, move.unit_price,
-                        self.company.currency, round=False)
-                move_unit_price = Uom.compute_price(
-                    move.uom, move_unit_price, move.product.default_uom)
-            else:
-                move_unit_price = move.cost_price or 0
-            cost_price += move_unit_price * Decimal(str(move_qty))
+            cost_price += move.get_cost_price() * Decimal(str(move_qty))
 
             move_qty = Uom.compute_qty(self.product.default_uom, move_qty,
                     move.uom, round=False)
             move.fifo_quantity = (move.fifo_quantity or 0.0) + move_qty
+            # Due to float, the fifo quantity result can exceed the quantity.
+            assert move.quantity >= move.fifo_quantity - move.uom.rounding
+            move.fifo_quantity = min(move.fifo_quantity, move.quantity)
             to_save.append(move)
-        if to_save:
-            # TODO save in do method when product change
-            self.__class__.save(to_save)
 
         if consumed_qty:
             cost_price = cost_price / Decimal(str(consumed_qty))
@@ -107,21 +97,18 @@ class Move(metaclass=PoolMeta):
                 'cost_price', **self._cost_price_pattern)
 
         # Compute average cost price
-        unit_price = self.unit_price
-        self.unit_price = Uom.compute_price(
-            self.product.default_uom, cost_price, self.uom)
-        average_cost_price = self._compute_product_cost_price('out')
-        self.unit_price = unit_price
+        average_cost_price = self._compute_product_cost_price(
+            'out', product_cost_price=cost_price)
 
         if cost_price:
             cost_price = round_price(cost_price)
         else:
             cost_price = average_cost_price
-        return cost_price, average_cost_price
+        return cost_price, average_cost_price, to_save
 
     def _do(self):
-        cost_price = super(Move, self)._do()
-        if (self.from_location.type in ('supplier', 'production')
+        cost_price, to_save = super(Move, self)._do()
+        if (self.from_location.type != 'storage'
                 and self.to_location.type == 'storage'
                 and self.product.cost_price_method == 'fifo'):
             cost_price = self._compute_product_cost_price('in')
@@ -132,17 +119,18 @@ class Move(metaclass=PoolMeta):
         elif (self.from_location.type == 'storage'
                 and self.to_location.type != 'storage'
                 and self.product.cost_price_method == 'fifo'):
-            fifo_cost_price, cost_price = (
+            fifo_cost_price, cost_price, moves = (
                 self._update_fifo_out_product_cost_price())
-            if self.cost_price is None:
+            if self.cost_price_required and self.cost_price is None:
                 self.cost_price = fifo_cost_price
-        return cost_price
+            to_save.extend(moves)
+        return cost_price, to_save
 
     @classmethod
     @ModelView.button
-    @Workflow.transition('cancel')
+    @Workflow.transition('cancelled')
     def cancel(cls, moves):
-        pass
+        super().cancel(moves)
 
     @classmethod
     def delete(cls, moves):
@@ -155,3 +143,12 @@ class Move(metaclass=PoolMeta):
                 gettext('product_cost_fifo.msg_move_delete_fifo',
                     move=fifo_moves[0].rec_name))
         super(Move, cls).delete(moves)
+
+    @classmethod
+    def copy(cls, moves, default=None):
+        if default is None:
+            default = {}
+        else:
+            default = default.copy()
+        default.setdefault('fifo_quantity', cls.default_fifo_quantity())
+        return super().copy(moves, default=default)
